require 'framework/common/tokenencoder'
require 'framework/common/utilities'
require 'framework/mixins/notifier'
require 'framework/lexer'

require 'strscan'


module Mondrian

    module TokenStringInternals

        def self.append_features(receiver)
            super
            
            receiver.extend(ClassMethods)
        end
        
        
        def lexer
            self.class.lexer
        end
    
        
        module ClassMethods
            attr_reader :lexer
        
            private
            def use_lexer(lexer_class)
                @lexer = lexer_class.new
        
                @lexer.tokens.each { |symbol|
                    module_eval <<-fin
                        def each_#{symbol}(text_match = nil, &block)
                            each(:#{symbol}, text_match, &block)
                        end

            
                        def next_#{symbol}(text_match = nil, start_position = 0)
                            self.next(:#{symbol}, text_match, start_position)
                        end
            
            
                        def previous_#{symbol}(text_match = nil, start_position = nil)
                            previous(:#{symbol}, text_match, start_position)
                        end
                    fin
                }
            end

        end

    end
        


    class TokenPointer
        attr_reader :position, :text, :token_type, :parent
        
        include Comparable
        
        def initialize(parent, position, length)
            @parent     = parent
            @position   = position
            
            @text       = parent.text[position...position + length]
            @token_type = parent.encoder.token_symbol(parent.tokens[position...position + length])
            
            @change = parent.changelog.length - 1
        end
        
        
        def <=>(object)
            if object.kind_of? self.class
                return position.send('<=>', object.position)
            elsif object.kind_of? Fixnum
                return position.send('<=>', object)
            end
            
            raise TypeError
        end
        
        
        def ==(object)
            if object.kind_of? String
                return text == object
            elsif object.kind_of? self.class
                return (object.text == text and object.token_type == token_type and object.position == position)
            elsif object.kind_of? Symbol
                return token_type == object
            elsif object.kind_of? Fixnum
                return position == object
            end
            
            super
        end
        
        
        def to_s
            text
        end
        
        
        def length
            text.length
        end
        
        
        def position
            update_position
            
            @position
        end
        
        
        # Corrects +position+ by looking at changes made to +parent+.
        
        def update_position
            return if @change == @parent.changelog.length - 1
        
            @parent.changelog[@change + 1...@parent.changelog.length].each { |change|
                @position += change[1] if @position >= change[0]
            }
            
            @change = @parent.changelog.length - 1
        end
        
        
        def deceased?
            !(self == @parent.token_at(position))
        end
        
        
        def assert_untainted
            raise "Object has expired!" if deceased?
        end


        def method_missing(method_symbol, *args)
            if method_symbol.to_s =~ /\A(next|previous)_?/
                case $&
                    when 'next'
                        return @parent.send(method_symbol, args[0], args[1], position + length) 
                    when 'previous'
                        return @parent.send(method_symbol, args[0], args[1], position - 1) 
                    when 'next_'
                        return @parent.send(method_symbol, args[0], position + length) 
                    when 'previous_'
                        return @parent.send(method_symbol, args[0], position - 1) 
                end
            end
            
            raise NoMethodError, method_symbol.to_s
        end
    end



    class TokenString
        include Enumerable
        include Notifier
        include TokenStringInternals
        
        attr_reader :text, :tokens, :encoder, :changelog
        
        
        def initialize(text = '')
            @encoder = TokenEncoder.new(lexer.tokens)

            @text        = ""
            @tokens      = ""
            
            @changelog = []
           
            insert_text(0, text)
        end
        
        
        def ==(object)
            if object.kind_of? String
                return @text == object
            elsif object.kind_of? self.class
                return object_id == object.object_id
            elsif object.kind_of? Symbol
                raise 'fix me' # FIXME
            end
            
            false
        end
        
        
        def text=(text)
            delete_text 0, @text.length
            insert_text 0, text
        end
        
        
        def length
            @text.length
        end
        
        
        def insert_text(index, new_text)
            @text[index, 0]   = new_text
            @tokens[index, 0] = @encoder.null_pad new_text.length
            
            @changelog << [ index, new_text.length ]

            lex_text(index, new_text.length)
        end
        
        
        def delete_text(index, length)
            @text[index, length]   = ''
            @tokens[index, length] = ''
        
            @changelog << [ index, -length ]

            lex_text(index)
        end


        def lex_text(index, length = 0)
            lex_start         = get_lex_start(index)
            context           = get_context(lex_start)
            last_change       = index + length - 1
            lex_end           = lex_start
            matched_pos       = 0
            encoded_token     = nil
            last_was_changed  = false
            force_continue    = true
            
            scanner     = StringScanner.new(@text)
            scanner.pos = lex_start
            
            while !scanner.eos? and (last_was_changed or force_continue)
                matched_pos = scanner.pos
            
                token_type, token_text = lexer.next_token(scanner, context)
                
                encoded_token = @encoder.encode_token(token_type, token_text.length)
                
                last_was_changed = true
                
                if scanner.pos > last_change and @tokens.index(encoder.match_any, matched_pos) == matched_pos
                    if $& == encoded_token and @text.index(token_text, matched_pos) == matched_pos
                        last_was_changed = false
                    end
                end

                if last_was_changed
                    @tokens[matched_pos..scanner.pos - 1] = encoded_token
                    
                    lex_end = scanner.pos - 1
                end
                
                force_continue = context.update!(token_text, token_type)
            end
            
            (lex_start..lex_end)
        end
        
        
        def to_s
            text
        end
        
        
        def token_at(index)
            return TokenPointer.new(self, $~.begin(0), $&.length) if @tokens.rindex(@encoder.match_any, index)

            nil                
        end
        
        
        def each(token_type = nil, text_match = nil)
            text_match  = regexp_from_string(text_match) if text_match and text_match.kind_of?(String)
            token_match = token_type ? @encoder.match(token_type) : @encoder.match_any
            
            scanner = StringScanner.new(@tokens)
            
            while scanner.skip_until(token_match)
                if not text_match or @text[scanner.pos - scanner.matched_size...scanner.pos] =~ text_match
                    yield TokenPointer.new(self, scanner.pos - scanner.matched_size, scanner.matched_size)
                end
            end
        end
        
        
        def next(token_type = nil, text_match = nil, start_position = 0)
            return nil if start_position < 0 or start_position >= @tokens.length
        
            text_match  = regexp_from_string(text_match) if text_match and text_match.kind_of?(String)
            token_match = token_type ? @encoder.match(token_type) : @encoder.match_any
            
            scanner = StringScanner.new(@tokens)
            scanner.pos = start_position
            
            while scanner.skip_until(token_match)
                if not text_match or @text[scanner.pos - scanner.matched_size...scanner.pos] =~ text_match
                    return TokenPointer.new(self, scanner.pos - scanner.matched_size, scanner.matched_size)
                end
            end
            
            nil
        end
        
        
        # FIXME: needs optimization
        def previous(token_type = nil, text_match = nil, start_position = nil)
            start_position = @tokens.length unless start_position
        
            return nil if start_position <= 0 or start_position > @tokens.length
            
            match = regexp_from_string(match) if text_match and text_match.kind_of?(String)

            token_index = start_position

            while token_index and token_index >= 0
                token_index = @tokens.rindex(@encoder.match(token_type), token_index)
                
                if token_index
                    encoded_token = $&

                    if not text_match or @text[token_index...token_index + encoded_token.length] =~ text_match
                        return TokenPointer.new(self, token_index, encoded_token.length)
                    end
                    
                    token_index -= 1
                end
            end
            
            return nil
        end
        
        
        protected
        def get_lex_start(insert_index)
            insert_index
        end
        
        
        protected
        def get_context(index)
            i = @tokens.rindex(@encoder.match_any, index - 1)
            
            last = i ? @text[i, $&.length] : nil
            last_token = last ? @encoder.token_symbol($&) : nil

            LexerContext.new(last, last_token)
        end

    end

end